package utils

import (
	"bytes"
	"crypto/tls"
	"crypto/x509"
	"encoding/json"
	"errors"
	"fmt"
	"io"
	"io/ioutil"
	"net"
	"net/http"
	"time"
)

const (
	IngestAPI           = "https://deepfence-fetcher:8006/df-api/ingest?doc_type=cve"
	IngestScanStatusAPI = "https://deepfence-fetcher:8006/df-api/ingest?doc_type=cve-scan"
	MaskedCVEAPI        = "https://deepfence-fetcher:8006/df-api/masked-cve-id"
)

func buildHttpClient() (*http.Client, error) {
	// Set up our own certificate pool
	tlsConfig := &tls.Config{RootCAs: x509.NewCertPool(), InsecureSkipVerify: true}
	client := &http.Client{
		Transport: &http.Transport{
			TLSClientConfig:     tlsConfig,
			DisableKeepAlives:   false,
			MaxIdleConnsPerHost: 1024,
			DialContext: (&net.Dialer{
				Timeout:   15 * time.Minute,
				KeepAlive: 15 * time.Minute,
			}).DialContext,
			TLSHandshakeTimeout:   10 * time.Second,
			ResponseHeaderTimeout: 5 * time.Minute,
		},
		Timeout: 15 * time.Minute,
	}
	return client, nil
}

func GetCurrentlyMaskedCveIds(nodeId, nodeType string) ([]string, error) {
	var currentlyMaskedCveIds []string

	httpClient, err := buildHttpClient()
	if err != nil {
		return currentlyMaskedCveIds, err
	}

	jsonPayload, err := json.Marshal(map[string]string{"node_id": nodeId, "node_type": nodeType})
	if err != nil {
		return currentlyMaskedCveIds, err
	}
	httpReq, err := http.NewRequest("POST", MaskedCVEAPI, bytes.NewBuffer(jsonPayload))
	if err != nil {
		return currentlyMaskedCveIds, err
	}
	httpReq.Close = true
	resp, err := httpClient.Do(httpReq)
	if err != nil {
		return currentlyMaskedCveIds, err
	}
	defer resp.Body.Close()
	if resp.StatusCode == 200 {
		maskedCveIdStr, err := ioutil.ReadAll(resp.Body)
		if err != nil {
			return currentlyMaskedCveIds, err
		}
		err = json.Unmarshal(maskedCveIdStr, &currentlyMaskedCveIds)
		if err != nil {
			return currentlyMaskedCveIds, err
		}
	}
	return currentlyMaskedCveIds, nil
}

func PostRequest(requestUrl string, postReader io.Reader, httpHeader http.Header) error {
	retryCount := 0
	httpClient, err := buildHttpClient()
	if err != nil {
		return err
	}
	for {
		httpReq, err := http.NewRequest("POST", requestUrl, postReader)
		if err != nil {
			return err
		}
		httpReq.Close = true
		httpReq.Header = httpHeader
		resp, err := httpClient.Do(httpReq)
		if err != nil {
			return err
		}
		if resp.StatusCode == 200 {
			resp.Body.Close()
			break
		} else {
			if retryCount > 2 {
				errMsg := fmt.Sprintf("Unable to complete request. Got %d ", resp.StatusCode)
				resp.Body.Close()
				return errors.New(errMsg)
			}
			resp.Body.Close()
			retryCount += 1
			time.Sleep(5 * time.Second)
		}
	}
	return nil
}
